import torch
from torch import Tensor
from torch.linalg import vector_norm
from torch.nn import Module

from haplo.nicer_transform import PrecomputedUnnormalizePhaseAmplitudes


class PlusOneChiSquaredStatisticMetric(Module):
    def forward(self, output: torch.Tensor, target: torch.Tensor):
        unnormalize_phase_amplitudes = PrecomputedUnnormalizePhaseAmplitudes()
        observed = unnormalize_phase_amplitudes(output.type(torch.float64)) + 1.0
        expected = unnormalize_phase_amplitudes(target.type(torch.float64)) + 1.0
        chi_squared_statistic_f64 = torch.mean(torch.sum(((observed - expected) ** 2) / expected, dim=1))
        chi_squared_statistic = chi_squared_statistic_f64.type(torch.float32)
        return chi_squared_statistic


class PlusOneBeforeUnnormalizationChiSquaredStatisticMetric(Module):
    def forward(self, output: torch.Tensor, target: torch.Tensor):
        unnormalize_phase_amplitudes = PrecomputedUnnormalizePhaseAmplitudes()
        observed = unnormalize_phase_amplitudes(output.type(torch.float64) + 1.0)
        expected = unnormalize_phase_amplitudes(target.type(torch.float64) + 1.0)
        chi_squared_statistic_f64 = torch.mean(torch.sum(((observed - expected) ** 2) / expected, dim=1))
        chi_squared_statistic = chi_squared_statistic_f64.type(torch.float32)
        return chi_squared_statistic


def norm_based_gradient_clip(gradient_tensor: Tensor) -> Tensor:
    return gradient_tensor / torch.maximum(torch.linalg.vector_norm(gradient_tensor), torch.tensor(1.0))


import mmap
from pathlib import Path
from typing import Optional, Callable, List

import numpy as np
import polars as pl
import pyarrow
from pyarrow import feather, MemoryMappedFile
from tenacity import wait_random, retry, wait_exponential, stop_after_attempt, retry_if_exception_type
from torch.utils.data import Dataset, Subset

from haplo.data_column_name import DataColumnName


class NicerDataset(Dataset):
    def __init__(self, dataset_path: Path, parameters_transform: Optional[Callable] = None,
                 phase_amplitudes_transform: Optional[Callable] = None):
        self.dataset_path: Path = dataset_path
        self.parameters_transform: Callable = parameters_transform
        self.phase_amplitudes_transform: Callable = phase_amplitudes_transform
        self.database_uri = f'sqlite://{self.dataset_path}'
        # TODO: Quick hack. Should not being doing logic in init. Move this to factory method.
        count_data_frame = pl.read_database_uri(query='select count(1) from main', uri=self.database_uri)
        count_row = count_data_frame.row(0)
        count = count_row[0]
        self.length: int = count

    @classmethod
    def new(cls, dataset_path: Path, parameters_transform: Optional[Callable] = None,
            phase_amplitudes_transform: Optional[Callable] = None):
        instance = cls(dataset_path=dataset_path, parameters_transform=parameters_transform,
                       phase_amplitudes_transform=phase_amplitudes_transform)
        return instance

    def __len__(self):
        return self.length

    def __getitem__(self, index):
        row_index = index + 1  # The SQL database auto increments from 1, not 0.
        row = self.get_row_from_index(row_index)
        parameters = np.array(row[:11], dtype=np.float32)
        phase_amplitudes = np.array(row[11:], dtype=np.float32)
        if self.parameters_transform is not None:
            parameters = self.parameters_transform(parameters)
        if self.phase_amplitudes_transform is not None:
            phase_amplitudes = self.phase_amplitudes_transform(phase_amplitudes)
        return parameters, phase_amplitudes

    # TODO: This retry probably shouldn't be necessary at all. But even if it is, I should probably be catching
    #       only a specific kind of exception here (i.e., pyo3_runtime.PanicException).
    @retry(wait=wait_exponential(multiplier=1, min=1, max=10), stop=stop_after_attempt(10))
    def get_row_from_index(self, row_index):
        row_data_frame = pl.read_database_uri(query=rf'select * from main where ROWID = {row_index}',
                                              uri=self.database_uri)
        row = row_data_frame.row(0)
        return row


def split_into_train_validation_and_test_datasets(dataset: NicerDataset) -> (NicerDataset, NicerDataset, NicerDataset):
    length_10_percent = round(len(dataset) * 0.1)
    train_dataset = Subset(dataset, range(length_10_percent * 8))
    validation_dataset = Subset(dataset, range(length_10_percent * 8, length_10_percent * 9))
    test_dataset = Subset(dataset, range(length_10_percent * 9, len(dataset)))
    return train_dataset, validation_dataset, test_dataset


def split_dataset_into_fractional_datasets(dataset: NicerDataset, fractions: List[float]) -> List[NicerDataset]:
    assert np.isclose(np.sum(fractions), 1.0)
    fractional_datasets: List[NicerDataset] = []
    cumulative_fraction = 0
    previous_index = 0
    for fraction in fractions:
        cumulative_fraction += fraction
        if np.isclose(cumulative_fraction, 1.0):
            next_index = len(dataset)
        else:
            next_index = round(len(dataset) * cumulative_fraction)
        fractional_dataset: NicerDataset = Subset(dataset, range(previous_index, next_index))
        fractional_datasets.append(fractional_dataset)
        previous_index = next_index
    return fractional_datasets


def split_dataset_into_count_datasets(dataset: NicerDataset, counts: List[int]) -> List[NicerDataset]:
    assert np.sum(counts) < len(dataset)
    count_datasets: List[NicerDataset] = []
    next_index = 0
    previous_index = 0
    for count in counts:
        next_index += count
        count_dataset: NicerDataset = Subset(dataset, range(previous_index, next_index))
        count_datasets.append(count_dataset)
        previous_index = next_index
    count_datasets.append(Subset(dataset, range(previous_index, len(dataset))))
    return count_datasets



import re
import mmap
from pathlib import Path
from typing import TextIO, Dict, List

import polars as pl

from haplo.data_column_name import DataColumnName
from haplo.data_paths import constantinos_kalapotharakos_format_rotated_dataset_path, rotated_dataset_path, \
    constantinos_kalapotharakos_format_unrotated_dataset_path, unrotated_dataset_path


def constantinos_kalapotharakos_file_handle_to_sqlite(file_contents: bytes | mmap.mmap, output_file_path: Path
                                                      ):
    output_file_path.unlink(missing_ok=True)
    Path(str(output_file_path) + '-shm').unlink(missing_ok=True)
    Path(str(output_file_path) + '-wal').unlink(missing_ok=True)
    Path(str(output_file_path) + '-journal').unlink(missing_ok=True)
    value_iterator = re.finditer(rb"[^\s]+", file_contents)
    list_of_dictionaries: List[Dict] = []
    data_frame = pl.from_dicts([], schema={str(name): pl.Float32 for name in DataColumnName})
    count = 0
    while True:
        parameters = []
        try:
            parameters.append(float(next(value_iterator).group(0)))
        except StopIteration:
            break
        for _ in range(10):
            parameters.append(float(next(value_iterator).group(0)))
        _ = float(next(value_iterator).group(0))  # Likelihood in Constantinos' output which has no meaning here.
        phase_amplitudes = []
        for _ in range(64):
            phase_amplitudes.append(float(next(value_iterator).group(0)))
        row_values = parameters + phase_amplitudes
        row_dictionary = {str(name): value for name, value in zip(DataColumnName, row_values)}
        list_of_dictionaries.append(row_dictionary)
        count += 1
        if len(list_of_dictionaries) % 100000 == 0:
            print(f'{count}', flush=True)
            chunk_data_frame = pl.from_dicts(list_of_dictionaries, schema={str(name): pl.Float32 for name in DataColumnName})
            chunk_data_frame.write_database('main', f'sqlite:///{output_file_path}', if_exists='append')
            list_of_dictionaries = []
    chunk_data_frame = pl.from_dicts(list_of_dictionaries, schema={str(name): pl.Float32 for name in DataColumnName})
    chunk_data_frame.write_database('main', f'sqlite:///{output_file_path}', if_exists='append')


def arbitrary_constantinos_kalapotharakos_file_handle_to_polars(data_path: Path, columns_per_row: int
                                                                ) -> pl.DataFrame:
    with data_path.open() as file_handle:
        file_contents = get_memory_mapped_file_contents(file_handle)
        value_iterator = re.finditer(rb"[^\s]+", file_contents)
        list_of_dictionaries: List[Dict] = []
        data_frame = pl.from_dicts([], schema={str(index): pl.Float32 for index in range(columns_per_row)})
        count = 0
        while True:
            values = []
            try:
                values.append(float(next(value_iterator).group(0)))
            except StopIteration:
                break
            for _ in range(columns_per_row - 1):
                values.append(float(next(value_iterator).group(0)))
            row_dictionary = {str(index): value for index, value in zip(range(columns_per_row), values)}
            list_of_dictionaries.append(row_dictionary)
            count += 1
            if len(list_of_dictionaries) % 100000 == 0:
                print(f'{count}', flush=True)
                chunk_data_frame = pl.from_dicts(list_of_dictionaries,
                                                 schema={str(index): pl.Float32 for index in range(columns_per_row)})
                data_frame = data_frame.vstack(chunk_data_frame)
                list_of_dictionaries = []
        chunk_data_frame = pl.from_dicts(list_of_dictionaries,
                                         schema={str(index): pl.Float32 for index in range(columns_per_row)})
        data_frame = data_frame.vstack(chunk_data_frame)
        return data_frame


def get_memory_mapped_file_contents(file_handle: TextIO) -> mmap.mmap:
    file_fileno = file_handle.fileno()
    file_contents = mmap.mmap(file_fileno, 0, access=mmap.ACCESS_READ)
    return file_contents


def constantinos_kalapotharakos_format_file_to_sqlite(input_file_path: Path, output_file_path: Path) -> None:
    """
    Produces an SQLite database from a Constantinos Kalapotharakos format file. The expected input format includes
    11 parameters, 1 likelihood value (which is ignored), and 64 phase amplitude values for each entry.

    :param input_file_path: The Path to the Constantinos Kalapotharakos format file.
    :param output_file_path: The Path to the output SQLite database.
    """
    with input_file_path.open() as file_handle:
        file_contents = get_memory_mapped_file_contents(file_handle)
        constantinos_kalapotharakos_file_handle_to_sqlite(file_contents, output_file_path)


if __name__ == '__main__':
    constantinos_kalapotharakos_format_file_to_sqlite(
        Path('data/mcmc_vac_all_640m_A.dat'), Path('data/640m_rotated_parameters_and_phase_amplitudes.db'))



from enum import StrEnum


class DataColumnName(StrEnum):
    PARAMETER0 = 'parameter0'
    PARAMETER1 = 'parameter1'
    PARAMETER2 = 'parameter2'
    PARAMETER3 = 'parameter3'
    PARAMETER4 = 'parameter4'
    PARAMETER5 = 'parameter5'
    PARAMETER6 = 'parameter6'
    PARAMETER7 = 'parameter7'
    PARAMETER8 = 'parameter8'
    PARAMETER9 = 'parameter9'
    PARAMETER10 = 'parameter10'
    PHASE_AMPLITUDE0 = 'phase_amplitude0'
    PHASE_AMPLITUDE1 = 'phase_amplitude1'
    PHASE_AMPLITUDE2 = 'phase_amplitude2'
    PHASE_AMPLITUDE3 = 'phase_amplitude3'
    PHASE_AMPLITUDE4 = 'phase_amplitude4'
    PHASE_AMPLITUDE5 = 'phase_amplitude5'
    PHASE_AMPLITUDE6 = 'phase_amplitude6'
    PHASE_AMPLITUDE7 = 'phase_amplitude7'
    PHASE_AMPLITUDE8 = 'phase_amplitude8'
    PHASE_AMPLITUDE9 = 'phase_amplitude9'
    PHASE_AMPLITUDE10 = 'phase_amplitude10'
    PHASE_AMPLITUDE11 = 'phase_amplitude11'
    PHASE_AMPLITUDE12 = 'phase_amplitude12'
    PHASE_AMPLITUDE13 = 'phase_amplitude13'
    PHASE_AMPLITUDE14 = 'phase_amplitude14'
    PHASE_AMPLITUDE15 = 'phase_amplitude15'
    PHASE_AMPLITUDE16 = 'phase_amplitude16'
    PHASE_AMPLITUDE17 = 'phase_amplitude17'
    PHASE_AMPLITUDE18 = 'phase_amplitude18'
    PHASE_AMPLITUDE19 = 'phase_amplitude19'
    PHASE_AMPLITUDE20 = 'phase_amplitude20'
    PHASE_AMPLITUDE21 = 'phase_amplitude21'
    PHASE_AMPLITUDE22 = 'phase_amplitude22'
    PHASE_AMPLITUDE23 = 'phase_amplitude23'
    PHASE_AMPLITUDE24 = 'phase_amplitude24'
    PHASE_AMPLITUDE25 = 'phase_amplitude25'
    PHASE_AMPLITUDE26 = 'phase_amplitude26'
    PHASE_AMPLITUDE27 = 'phase_amplitude27'
    PHASE_AMPLITUDE28 = 'phase_amplitude28'
    PHASE_AMPLITUDE29 = 'phase_amplitude29'
    PHASE_AMPLITUDE30 = 'phase_amplitude30'
    PHASE_AMPLITUDE31 = 'phase_amplitude31'
    PHASE_AMPLITUDE32 = 'phase_amplitude32'
    PHASE_AMPLITUDE33 = 'phase_amplitude33'
    PHASE_AMPLITUDE34 = 'phase_amplitude34'
    PHASE_AMPLITUDE35 = 'phase_amplitude35'
    PHASE_AMPLITUDE36 = 'phase_amplitude36'
    PHASE_AMPLITUDE37 = 'phase_amplitude37'
    PHASE_AMPLITUDE38 = 'phase_amplitude38'
    PHASE_AMPLITUDE39 = 'phase_amplitude39'
    PHASE_AMPLITUDE40 = 'phase_amplitude40'
    PHASE_AMPLITUDE41 = 'phase_amplitude41'
    PHASE_AMPLITUDE42 = 'phase_amplitude42'
    PHASE_AMPLITUDE43 = 'phase_amplitude43'
    PHASE_AMPLITUDE44 = 'phase_amplitude44'
    PHASE_AMPLITUDE45 = 'phase_amplitude45'
    PHASE_AMPLITUDE46 = 'phase_amplitude46'
    PHASE_AMPLITUDE47 = 'phase_amplitude47'
    PHASE_AMPLITUDE48 = 'phase_amplitude48'
    PHASE_AMPLITUDE49 = 'phase_amplitude49'
    PHASE_AMPLITUDE50 = 'phase_amplitude50'
    PHASE_AMPLITUDE51 = 'phase_amplitude51'
    PHASE_AMPLITUDE52 = 'phase_amplitude52'
    PHASE_AMPLITUDE53 = 'phase_amplitude53'
    PHASE_AMPLITUDE54 = 'phase_amplitude54'
    PHASE_AMPLITUDE55 = 'phase_amplitude55'
    PHASE_AMPLITUDE56 = 'phase_amplitude56'
    PHASE_AMPLITUDE57 = 'phase_amplitude57'
    PHASE_AMPLITUDE58 = 'phase_amplitude58'
    PHASE_AMPLITUDE59 = 'phase_amplitude59'
    PHASE_AMPLITUDE60 = 'phase_amplitude60'
    PHASE_AMPLITUDE61 = 'phase_amplitude61'
    PHASE_AMPLITUDE62 = 'phase_amplitude62'
    PHASE_AMPLITUDE63 = 'phase_amplitude63'




import multiprocessing
import multiprocessing
import os
from pathlib import Path
from typing import Callable, List, Dict, Any

import numpy as np
import stringcase
import torch
import wandb as wandb
from torch import multiprocessing, Tensor
from torch.distributed import init_process_group, destroy_process_group, Backend
from torch.nn import Module
from torch.nn.parallel import DistributedDataParallel
from torch.optim import AdamW, Optimizer
from torch.types import Device
from torch.utils.data import DataLoader, DistributedSampler, Dataset

from haplo.data_paths import move_path_to_nvme
from haplo.losses import PlusOneChiSquaredStatisticMetric, PlusOneBeforeUnnormalizationChiSquaredStatisticMetric, \
    norm_based_gradient_clip
from haplo.models import Cura
from haplo.nicer_dataset import NicerDataset, split_dataset_into_fractional_datasets
from haplo.nicer_transform import PrecomputedNormalizeParameters, PrecomputedNormalizePhaseAmplitudes
from haplo.wandb_liaison import wandb_set_run_name, wandb_init, wandb_log, wandb_commit, \
    wandb_log_hyperparameter_dictionary


def ddp_setup():
    if torch.cuda.is_available():
        distributed_back_end = Backend.NCCL
    else:
        distributed_back_end = Backend.GLOO
    distributed_back_end = Backend.GLOO
    if 'RANK' not in os.environ:
        # The script was not called with `torchrun` and environment variables need to be set manually.
        os.environ['RANK'] = str(0)
        os.environ['LOCAL_RANK'] = str(0)
        os.environ['WORLD_SIZE'] = str(1)
        os.environ["MASTER_ADDR"] = "localhost"
        os.environ["MASTER_PORT"] = "35728"
    init_process_group(backend=distributed_back_end)


def default_train_session():
    train_dataset_path = Path('data/50m_rotated_parameters_and_phase_amplitudes.db')
    train_dataset_path = move_path_to_nvme(train_dataset_path)
    full_train_dataset = NicerDataset.new(
        dataset_path=train_dataset_path,
        parameters_transform=PrecomputedNormalizeParameters(),
        phase_amplitudes_transform=PrecomputedNormalizePhaseAmplitudes())
    train_dataset, validation_dataset, test_dataset = split_dataset_into_fractional_datasets(full_train_dataset,
                                                                                             [0.8, 0.1, 0.1])
    model = Cura()
    for parameter in model.parameters():
        parameter.register_hook(norm_based_gradient_clip)
    loss_function = PlusOneBeforeUnnormalizationChiSquaredStatisticMetric()
    metric_functions = [PlusOneChiSquaredStatisticMetric(), PlusOneBeforeUnnormalizationChiSquaredStatisticMetric()]
    learning_rate = 1e-4
    optimizer = AdamW(model.parameters(), lr=learning_rate, eps=1e-7)
    batch_size_per_device = 100
    cycles_to_run = 5000
    model_name = type(model).__name__
    run_notes = f"{model_name}_old_chi_squared_loss_shuffled_50m_dataloader_shuffled_bs_{batch_size_per_device}" \
               f"_copy_on_transform_train_and_val_from_same_corrected2_val_calc_adamw_grad_norm_clip_1_node" \
               f"_spawn_w3_no_wd_sqlite_db"
    hyperparameter_log_dictionary = {'learning_rate': learning_rate}
    train_session(train_dataset, validation_dataset, model, loss_function, metric_functions, optimizer,
                  batch_size_per_device, cycles_to_run, run_notes, wandb_project='haplo', wandb_entity='ramjet',
                  hyperparameter_log_dictionary=hyperparameter_log_dictionary)


def train_session(train_dataset: Dataset, validation_dataset: Dataset, model: Module, loss_function: Module,
                  metric_functions: List[Module], optimizer: Optimizer, batch_size_per_device: int, cycles_to_run: int,
                  run_notes: str, wandb_project: str, wandb_entity: str,
                  hyperparameter_log_dictionary: Dict[str, Any] | None = None):
    if hyperparameter_log_dictionary is None:
        hyperparameter_log_dictionary = {}
    print('Starting training...')
    print('Starting process spawning...')
    torch.multiprocessing.set_start_method('spawn')
    print('Starting DDP setup...')
    ddp_setup()
    process_rank = int(os.environ['RANK'])
    print(f'{process_rank}: Starting wandb...')
    wandb_init(process_rank=process_rank, project=wandb_project, entity=wandb_entity,
               settings=wandb.Settings(start_method='fork'))
    wandb_log_hyperparameter_dictionary(hyperparameter_log_dictionary, process_rank=process_rank)
    cpu_count = multiprocessing.cpu_count()
    gpu_count = torch.cuda.device_count()
    print(f'GPUs: {gpu_count}')

    if torch.cuda.is_available():
        local_rank = int(os.environ['LOCAL_RANK'])
        network_device = torch.device(f'cuda:{local_rank}')
        loss_device = network_device
    else:
        network_device = torch.device('cpu')
        loss_device = network_device

    print(f'{process_rank}: Moving model to device...')
    model = model.to(network_device, non_blocking=True)
    if torch.cuda.is_available():
        local_rank = int(os.environ['LOCAL_RANK'])
        model = DistributedDataParallel(model, device_ids=[local_rank])
    else:
        model = DistributedDataParallel(model)

    print(f'{process_rank}: Loading dataset...')
    train_dataloader = DataLoader(train_dataset, batch_size=batch_size_per_device, num_workers=3, pin_memory=True,
                                  persistent_workers=True, prefetch_factor=10, shuffle=False,
                                  sampler=DistributedSampler(train_dataset))
    validation_dataloader = DataLoader(validation_dataset, batch_size=batch_size_per_device, num_workers=3,
                                       pin_memory=True, persistent_workers=True, prefetch_factor=10, shuffle=False,
                                       sampler=DistributedSampler(validation_dataset))

    wandb_set_run_name(run_notes, process_rank=process_rank)

    print(f'{process_rank}: Starting training loop...')
    for cycle in range(cycles_to_run):
        print(f"Epoch {cycle}\n-------------------------------")
        train_loop(train_dataloader, model, loss_function, optimizer, network_device=network_device,
                   loss_device=loss_device, cycle=cycle, metric_functions=metric_functions, process_rank=process_rank)
        loop_test(validation_dataloader, model, loss_function, network_device=network_device, loss_device=loss_device,
                  cycle=cycle, metric_functions=metric_functions, process_rank=process_rank)
        save_model(model, process_rank=process_rank)
        wandb_log('epoch', cycle, process_rank=process_rank)
        wandb_log('cycle', cycle, process_rank=process_rank)
        wandb_commit(process_rank=process_rank)
    print("Done!")

    destroy_process_group()


def save_model(model: Module, process_rank: int):
    if process_rank == 0:
        torch.save(model.state_dict(), Path(f'sessions/{wandb.run.id}_latest_model.pt'))


def train_loop(dataloader: DataLoader, model: Module, loss_function: Callable[[Tensor, Tensor], Tensor],
               optimizer: Optimizer, network_device: Device, loss_device: Device, cycle: int,
               metric_functions: List[Callable[[Tensor, Tensor], Tensor]], process_rank: int):
    number_of_batches = len(dataloader)
    model.train()
    total_cycle_loss = 0
    metric_totals = np.zeros(shape=[len(metric_functions)])
    assert isinstance(dataloader.sampler, DistributedSampler)
    dataloader.sampler.set_epoch(cycle)
    for batch, (parameters, light_curves) in enumerate(dataloader):
        parameters = parameters.to(network_device, non_blocking=True)
        light_curves = light_curves.to(loss_device, non_blocking=True)
        predicted_light_curves = model(parameters)
        loss = loss_function(predicted_light_curves.to(loss_device, non_blocking=True), light_curves).to(network_device,
                                                                                                         non_blocking=True)
        for metric_function_index, metric_function in enumerate(metric_functions):
            batch_metric_value = metric_function(predicted_light_curves.to(loss_device, non_blocking=True),
                                                 light_curves).item()
            metric_totals[metric_function_index] += batch_metric_value
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if batch % 1 == 0:
            loss_value, current = loss.item(), (batch + 1) * len(parameters)
            total_cycle_loss += loss_value
            print(f"loss: {loss_value:>7f}  [{current:>5d}/{len(dataloader.sampler):>5d}]", flush=True)
    wandb_log('loss', total_cycle_loss / number_of_batches, process_rank=process_rank)
    cycle_metric_values = metric_totals / number_of_batches
    for metric_function_index, metric_function in enumerate(metric_functions):
        wandb_log(f'{get_metric_name(metric_function)}', cycle_metric_values[metric_function_index],
                  process_rank=process_rank)


def get_metric_name(metric_function):
    metric_name = type(metric_function).__name__
    metric_name = stringcase.snakecase(metric_name)
    metric_name = metric_name.replace('_metric', '').replace('_loss', '')
    return metric_name


def loop_test(dataloader: DataLoader, model: Module, loss_function: Callable[[Tensor, Tensor], Tensor],
              network_device: Device, loss_device: Device, cycle: int,
              metric_functions: List[Callable[[Tensor, Tensor], Tensor]], process_rank: int):
    number_of_batches = len(dataloader)
    total_cycle_loss = 0
    metric_totals = np.zeros(shape=[len(metric_functions)])
    model.eval()
    assert isinstance(dataloader.sampler, DistributedSampler)
    dataloader.sampler.set_epoch(cycle)
    with torch.no_grad():
        for parameters, light_curves in dataloader:
            parameters = parameters.to(network_device, non_blocking=True)
            light_curves = light_curves.to(loss_device, non_blocking=True)
            predicted_light_curves = model(parameters)
            total_cycle_loss += loss_function(predicted_light_curves.to(loss_device, non_blocking=True), light_curves
                                              ).to(network_device, non_blocking=True).item()
            for metric_function_index, metric_function in enumerate(metric_functions):
                batch_metric_value = metric_function(predicted_light_curves.to(loss_device, non_blocking=True),
                                                     light_curves).item()
                metric_totals[metric_function_index] += batch_metric_value

    cycle_loss = total_cycle_loss / number_of_batches
    print(f"Test Error: \nAvg loss: {cycle_loss:>8f} \n", flush=True)
    wandb_log('val_plus_one_chi_squared_statistic', cycle_loss, process_rank=process_rank)
    cycle_metric_values = metric_totals / number_of_batches
    for metric_function_index, metric_function in enumerate(metric_functions):
        wandb_log(f'val_{get_metric_name(metric_function)}', cycle_metric_values[metric_function_index],
                  process_rank=process_rank)


if __name__ == '__main__':
    default_train_session()



import torch
from torch.nn import Module, DataParallel
from torch.utils.data import DataLoader

from haplo.data_paths import rotated_dataset_path, unrotated_dataset_path
from haplo.losses import PlusOneChiSquaredStatisticMetric
from haplo.models import LiraTraditionalShape8xWidthWithNoDoNoBn
from haplo.nicer_dataset import NicerDataset, split_dataset_into_fractional_datasets
from haplo.nicer_transform import PrecomputedNormalizeParameters, PrecomputedNormalizePhaseAmplitudes


def infer_session():
    if torch.backends.mps.is_available():
        device = torch.device('mps')
    elif torch.cuda.is_available():
        device = torch.device('cuda')
    else:
        device = torch.device('cpu')

    evaluation_dataset = NicerDataset.new(
        dataset_path=unrotated_dataset_path,
        parameters_transform=PrecomputedNormalizeParameters(),
        phase_amplitudes_transform=PrecomputedNormalizePhaseAmplitudes())
    validation_dataset, test_dataset = split_dataset_into_fractional_datasets(evaluation_dataset, [0.5, 0.5])

    batch_size = 1000

    test_dataloader = DataLoader(test_dataset, batch_size=batch_size)

    model = DataParallel(LiraTraditionalShape8xWidthWithNoDoNoBn())
    model = model.to(device)
    model.load_state_dict(torch.load('latest_model.pt'))
    model.eval()
    loss_function = PlusOneChiSquaredStatisticMetric()

    loop_test(test_dataloader, model, loss_function, device=device)


def loop_test(dataloader, model_: Module, loss_fn, device):
    num_batches = len(dataloader)
    test_loss, correct = 0, 0

    with torch.no_grad():
        for X, y in dataloader:
            X = X.to(device)
            y = y
            pred = model_(X)
            test_loss += loss_fn(pred.to('cpu'), y).to(device).item()

    test_loss /= num_batches
    print(f"Test Error: \nAvg loss: {test_loss:>8f} \n")


if __name__ == '__main__':
    infer_session()
